# -*- coding: utf-8 -*-
import itertools
import logging
import re
from codecs import open
from collections import namedtuple
from glob import glob
from os.path import basename

import pendulum

from tsstats import events
from tsstats.client import Clients

re_log_filename = re.compile(r'ts3server_(?P<date>\d{4}-\d\d-\d\d)'
                             r'__(?P<time>\d\d_\d\d_\d\d.\d+)_(?P<sid>\d).log')
re_log_entry = re.compile(r'(?P<timestamp>\d{4}-\d\d-\d\d\ \d\d:\d\d:\d\d.\d+)'
                          r'\|\ *(?P<level>\w+)\ *\|\ *(?P<component>\w+)\ *'
                          r'\|\ *(?P<sid>\d+)\ *\|\ *(?P<message>.*)')
re_dis_connect = re.compile(
    r"client (?P<action>(dis)?connected) '(?P<nick>.*)'\(id:(?P<clid>\d+)\)")
re_disconnect_invoker = re.compile(
    r'invokername=(.*)\ invokeruid=(.*)\ reasonmsg'
)

TimedLog = namedtuple('TimedLog', ['path', 'timestamp'])
Server = namedtuple('Server', ['sid', 'clients'])

logger = logging.getLogger('tsstats')


def _bundle_logs(logs):
    '''
    Bundle `logs` by virtualserver-id
    and sort by timestamp from filename (if exists)

    :param logs: list of paths to logfiles
    :type logs: list

    :return: `logs` bundled by virtualserver-id and sorted by timestamp
    :rtype: dict{str: [TimedLog]}
    '''
    vserver_logfiles = {}  # sid: [/path/to/log1, ..., /path/to/logn]
    for log in logs:
        # try to get date and sid from filename
        match = re_log_filename.match(basename(log))
        if match:
            match = match.groupdict()
            timestamp = pendulum.parse('{0} {1}'.format(
                match['date'], match['time'].replace('_', ':'))
            )
            tl = TimedLog(log, timestamp)
            sid = match['sid']
            if sid in vserver_logfiles:
                # if already exists, keep list sorted by timestamp
                vserver_logfiles[sid].append(tl)
                vserver_logfiles[sid] =\
                    sorted(vserver_logfiles[sid],
                           key=lambda tl: tl.timestamp)
            else:
                # if not exists, just create a list
                vserver_logfiles[sid] = [tl]
        else:
            # fallback to plain sorting
            vserver_logfiles.setdefault('', [])\
                .append(TimedLog(log, None))
            vserver_logfiles[''] =\
                sorted(vserver_logfiles[''],
                       key=lambda tl: tl.path)
    return vserver_logfiles


def _parse_line(line):
    '''
    Parse events from a single line

    :param line: line to parse events from
    :type line: str

    :return: parsed events
    :rtype list
    '''
    parsed_events = []
    match = re_log_entry.match(line)
    if not match:
        logger.debug('No match: "%s"', line)
        return []
    match = match.groupdict()
    logdatetime = pendulum.parse(match['timestamp'])
    message = match['message']
    if message.startswith('client'):
        match = re_dis_connect.match(message)
        if not match:
            logger.debug('Unsupported client action: "%s"', message)
            return []
        nick, clid = match.group('nick'), match.group('clid')

        parsed_events.append(events.nick(logdatetime, clid, nick))

        action = match.group('action')
        if action == 'connected':
            parsed_events.append(events.connect(logdatetime, clid))
        elif action == 'disconnected':
            parsed_events.append(events.disconnect(logdatetime, clid))
            if 'invokeruid' in message:
                re_disconnect_data = re_disconnect_invoker.findall(
                    message)
                invokernick, invokeruid = re_disconnect_data[0]
                parsed_events.append(
                    events.nick(logdatetime, invokeruid, invokernick)
                )
                if 'bantime' in message:
                    parsed_events.append(
                        events.ban(logdatetime, invokeruid, clid)
                    )
                else:
                    parsed_events.append(
                        events.kick(logdatetime, invokeruid, clid)
                    )
    return parsed_events


def parse_logs(log_glob, ident_map=None, online_dc=True):
    '''
    Parse logs from `log_glob`

    :param log_glob: path to server-logs (supports globbing)
    :param ident_map: identmap used for Client-initializations

    :type log_glob: str
    :type ident_map: dict

    :return: list of servers
    :rtype: [tsstats.log.Server]
    '''
    for virtualserver_id, logs in _bundle_logs(glob(log_glob)).items():
        clients = Clients(ident_map)
        for index, log in enumerate(logs):
            with open(log.path, encoding='utf-8') as f:
                logger.debug('Started parsing of %s', f.name)
                # parse logfile line by line and filter lines without events
                events = filter(None, map(_parse_line, f))
                all_events = list(itertools.chain.from_iterable(events))
                # chain apply events to Client-obj
                clients.apply_events(all_events)

                # find connected clients
                online_clients = list(
                    filter(lambda c: c.connected, clients.values())
                )

                if online_clients:
                    logger.debug(
                        'Some clients are still connected: %s', online_clients
                    )
                    if index == len(logs) - 1:
                        if online_dc:
                            logger.debug(
                                'Last log => disconnecting online clients'
                            )
                            # last iteration
                            # => disconnect online clients if desired
                            for online_client in online_clients:
                                online_client.disconnect(pendulum.now('UTC'))
                                online_client.connected += 1
                    else:
                        logger.warn(
                            'Server didn\'t disconnect all clients on shutdown'
                            ' or logfile is incorrectly named/corrupted (%s).'
                            ' Check debuglog for details',
                            f.name
                        )
                        logger.debug(
                            'Will handle this by disconnecting all clients on'
                            ' last event timestamp'
                        )
                        last_event_timestamp = all_events[-1].timestamp
                        logger.debug(
                            'Last event timestamp: %s', last_event_timestamp)
                        for online_client in online_clients:
                            online_client.disconnect(last_event_timestamp)

                logger.debug('Finished parsing of %s', f.name)
        if len(clients) >= 1:
            # assemble Server-obj and yield
            yield Server(virtualserver_id, clients)



# -*- coding: utf-8 -*-
import pytest

from tsstats.client import Client, Clients
from tsstats.log import parse_logs
from tsstats.utils import transform_pretty_identmap


@pytest.fixture(scope='module')
def identmap_clients():
    clients = Clients({
        '1': '2',
        '5': '2',
        'UID1': 'UID2',
        'UID5': 'UID2'
    })
    cl = Client('2', 'Client2')
    uidcl = Client('UID2', 'Client2++')
    clients += cl
    clients += uidcl
    return (clients, cl, uidcl)


def test_ident_map(identmap_clients):
    clients, cl, uidcl = identmap_clients
    assert clients['1'] == cl
    assert clients['5'] == cl
    assert clients['UID1'] == uidcl
    assert clients['UID5'] == uidcl


@pytest.mark.parametrize('test_input,expected', [
    (
        [
            {'primary_id': '1', 'alternate_ids': ['3', '6']},
            {'primary_id': '4', 'alternate_ids': ['9', '42', '23']}
        ],
        (('3', '1'), ('6', '1'), ('9', '4'), ('42', '4'), ('23', '4'))
    ),
    (
        [
            {'name': 'Friend 1', 'primary_id': '2', 'alternate_ids': ['4']},
            {
                'name': 'Friend 3',
                'primary_id': '8',
                'alternate_ids': ['9', '14']
            }
        ],
        (('4', '2'), ('9', '8'), ('14', '8'))
    )
])
def test_transform_pretty_identmap(test_input, expected):
    transformed_identmap = transform_pretty_identmap(test_input)
    for alternate, primary in expected:
        assert transformed_identmap[alternate] == primary


def test_ident_map_wrong_identifier():
    clients = list(parse_logs(
        'tsstats/tests/res/test.log.identmap_wrong_identifier', ident_map={
            '2': '1',
            '3': '1'
        }
    ))[0].clients
    client = clients.get('1')
    # assert client exists
    assert client
    # assert correct identifier
    assert client.identifier == '1'



# -*- coding: utf-8 -*-
import logging

import pendulum
import pytest
from bs4 import BeautifulSoup

from tsstats.log import parse_logs
from tsstats.template import render_servers
from tsstats.utils import filter_threshold, seconds_to_text, sort_clients

servers = list(parse_logs('tsstats/tests/res/test.log', online_dc=False))
servers[0] = servers[0]._replace(sid=1)  # add missing sid to server object
clients = servers[0].clients

logger = logging.getLogger('tsstats')


@pytest.fixture
def soup(output):
    render_servers(servers, output)
    return BeautifulSoup(open(output), 'html.parser')


def test_debug(output):
    logger.setLevel(logging.DEBUG)
    render_servers(servers, output)
    logger.setLevel(logging.INFO)
    soup = BeautifulSoup(open(output), 'html.parser')
    # check debug-label presence
    assert soup.find('nav').find('div', id='main-nav').find('span').text \
        == 'debug mode'
    for client_item in soup.find('ul', id='1.onlinetime').find_all('li'):
        nick = client_item.find('span').text
        # check for right identifier
        nick, encl_identifier = nick.split()
        identifier = encl_identifier.replace('(', '').replace(')', '')
        assert clients[identifier].nick == nick


def test_onlinetime(soup):
    items = soup.find('ul', id='1.onlinetime').find_all('li')
    assert len(items) == 2
    for item in items:
        nick, onlinetime = item.find_all('span')
        nick = nick.text
        onlinetime = onlinetime.text
        # find corresponding client-object
        client = list(filter(
            lambda c: c.nick == nick and c.onlinetime > pendulum.duration(),
            clients.values()
        ))
        # assert existence
        assert client
        client = client[0]
        # compare onlinetimes
        client_onlinetime_text = seconds_to_text(
            int(client.onlinetime.total_seconds())
        )
        assert onlinetime == client_onlinetime_text


def test_filter_threshold():
    sorted_clients = sort_clients(
        clients, lambda c: c.onlinetime.total_seconds())
    assert len(filter_threshold(sorted_clients, -1)) == len(sorted_clients)
    assert len(filter_threshold(sorted_clients, 20)) == 1
    assert len(filter_threshold(sorted_clients, 500)) == 0


def test_lastseen_relative(output):
    render_servers(servers, output, lastseen_relative=True)
    soup = BeautifulSoup(open(output), 'html.parser')
    assert soup.find('ul', id='1.onlinetime')\
        .select('div.hint--left')[0]['data-hint'] == \
        pendulum.datetime(2015, 5, 18).diff_for_humans()
    render_servers(servers, output, lastseen_relative=False)
    soup = BeautifulSoup(open(output), 'html.parser')
    assert soup.find('ul', id='1.onlinetime')\
        .select('div.hint--left')[0]['data-hint'] == \
        '05/18/15 15:54:38 UTC'



# -*- coding: utf-8 -*-
from os import remove

import pytest


@pytest.fixture
def output(request):
    output_path = 'tsstats/tests/res/output.html'

    def clean():
        remove(output_path)
    request.addfinalizer(clean)
    yield output_path



# -*- coding: utf-8 -*-
def sort_clients(clients, key_l):
    '''
    sort `clients` by `key`

    :param clients: clients to sort
    :param key_l: lambda/function returning the value of `key` for a client

    :type clients: tsstats.client.Clients
    :type key_l: function

    :return: sorted `clients`
    :rtype: list
    '''
    cl_data = [
        (client, key_l(client)) for client in clients.values()
        if key_l(client) > 0
    ]
    return sorted(cl_data, key=lambda data: data[1], reverse=True)


def seconds_to_text(seconds):
    '''
    convert `seconds` to a text-representation

    :param seconds: seconds to convert
    :type seconds: int

    :return: `seconds` as text-representation
    :rtype: str
    '''
    minutes, seconds = divmod(seconds, 60)
    hours, minutes = divmod(minutes, 60)
    hours = str(hours) + 'h ' if hours > 0 else ''
    minutes = str(minutes) + 'm ' if minutes > 0 else ''
    seconds = str(seconds) + 's' if seconds > 0 else ''
    return hours + minutes + seconds


def filter_threshold(clients, threshold):
    '''
    Filter clients by threshold

    :param clients: List of clients as returned by tsstats.utils.sort_clients
    :type clients: list

    :return: Clients matching given threshold
    :rtype: list
    '''
    return list(filter(lambda c: c[1] > threshold, clients))


def transform_pretty_identmap(pretty_identmap):
    '''
    Transforms a list of client ID mappings from a more descriptive format
    to the traditional format of alternative IDs to actual ID.

    :param pretty_identmap: ID mapping in "nice" form
    :type pretty_identmap: list

    :return: ID mapping in simple key/value pairs
    :rtype: dict
    '''

    final_identmap = {}
    for mapping in pretty_identmap:
        for alt_id in mapping['alternate_ids']:
            final_identmap[alt_id] = mapping['primary_id']
    return final_identmap



# -*- coding: utf-8 -*-
import argparse
import json
import logging
from os.path import join as pathjoin
from os.path import abspath, exists, isdir
from time import time

from tsstats import config
from tsstats.exceptions import InvalidConfiguration
from tsstats.log import parse_logs
from tsstats.logger import file_handler, stream_handler
from tsstats.template import render_servers
from tsstats.utils import transform_pretty_identmap

logger = logging.getLogger('tsstats')


def cli():
    parser = argparse.ArgumentParser(
        description='A simple Teamspeak stats-generator,'
        ' based solely on server-logs',
        argument_default=argparse.SUPPRESS
    )
    parser.add_argument(
        '-c', '--config',
        type=str, help='path to config'
    )
    parser.add_argument(
        '--idmap', type=str, help='path to id_map'
    )
    parser.add_argument(
        '-l', '--log',
        type=str, help='path to your logfile(s). '
        'pass a directory to use all logfiles inside it'
    )
    parser.add_argument(
        '-o', '--output',
        type=str, help='path to the output-file'
    )
    parser.add_argument(
        '-d', '--debug',
        help='debug mode', action='store_true'
    )
    parser.add_argument(
        '-ds', '--debugstdout',
        help='write debug output to stdout', action='store_true'
    )
    parser.add_argument(
        '-nod', '--noonlinedc',
        help='don\'t add connect until now to onlinetime',
        action='store_false', dest='onlinedc'
    )
    parser.add_argument(
        '-t', '--template',
        type=str, help='path to custom template'
    )
    parser.add_argument(
        '-dtf', '--datetimeformat',
        type=str, help='format of date/time-values (datetime.strftime)'
    )
    parser.add_argument(
        '-otth', '--onlinetimethreshold',
        type=int, help='threshold for displaying onlinetime (in seconds)'
    )
    parser.add_argument(
        '-lsa', '--lastseenabsolute',
        help='render last seen timestamp absolute (instead of relative)',
        action='store_false', dest='lastseenrelative'
    )
    options = parser.parse_args()
    if 'config' in options:
        configuration = config.load(options.config)
    else:
        configuration = config.load()
    for option, value in vars(options).items():
        configuration.set('General', option, str(value))
    main(configuration)


def main(configuration):
    start_time = time()
    # setup logging
    if configuration.getboolean('General', 'debug'):
        logger.setLevel(logging.DEBUG)
    if configuration.getboolean('General', 'debugstdout'):
        stream_handler.setLevel(logging.DEBUG)
    else:
        logger.addHandler(file_handler)

    # attach handlers
    logger.addHandler(stream_handler)

    idmap = configuration.get('General', 'idmap')
    if idmap:
        idmap = abspath(idmap)
        if not exists(idmap):
            logger.fatal('identmap not found (%s)', idmap)
        # read id_map
        identmap = json.load(open(idmap))
    else:
        identmap = None
    if isinstance(identmap, list):
        identmap = transform_pretty_identmap(identmap)

    log = configuration.get('General', 'log')
    if not log:
        raise InvalidConfiguration('log or output missing')
    if isdir(log):
        log = pathjoin(log, '*.log')

    servers = parse_logs(
        log, ident_map=identmap,
        online_dc=configuration.getboolean('General', 'onlinedc')
    )
    render_servers(
        sorted(servers, key=lambda s: s.sid),
        output=abspath(configuration.get('General', 'output')),
        template=configuration.get('General', 'template'),
        datetime_fmt=configuration.get('General', 'datetimeformat'),
        onlinetime_threshold=int(configuration.get(
            'General', 'onlinetimethreshold'
        )),
        lastseen_relative=configuration.getboolean(
            'General', 'lastseenrelative'
        )
    )
    logger.info('Finished after %s seconds', time() - start_time)


if __name__ == '__main__':
    cli()



# -*- coding: utf-8 -*-
from collections import namedtuple

Event = namedtuple(
    'Event', ['timestamp', 'identifier', 'action', 'arg', 'arg_is_client']
)


def nick(timestamp, identifier, nick):
    return Event(timestamp, identifier, 'set_nick', nick, arg_is_client=False)


def connect(timestamp, identifier):
    return Event(
        timestamp, identifier, 'connect', arg=timestamp, arg_is_client=False
    )


def disconnect(timestamp, identifier):
    return Event(
        timestamp, identifier, 'disconnect', arg=timestamp, arg_is_client=False
    )


def kick(timestamp, identifier, target_identifier):
    return Event(
        timestamp, identifier, 'kick', target_identifier, arg_is_client=True
    )


def ban(timestamp, identifier, target_identifier):
    return Event(
        timestamp, identifier, 'ban', target_identifier, arg_is_client=True
    )



# -*- coding: utf-8 -*-
import logging

# setup logger
logger = logging.getLogger('tsstats')
logger.setLevel(logging.INFO)

# define handlers
file_handler = logging.FileHandler('debug.txt', 'w', delay=True)
file_handler.setLevel(logging.DEBUG)

stream_handler = logging.StreamHandler()
stream_handler.setLevel(logging.INFO)



# -*- coding: utf-8 -*-
import pytest

from tsstats.config import load


@pytest.fixture
def config():
    return load()


def test_config(config):
    assert not config.getboolean('General', 'debug')
    assert config.getboolean('General', 'onlinedc')
    config.set('General', 'idmap', 'tsstats/tests/res/id_map.json')
    assert config.get('General', 'idmap') ==\
        'tsstats/tests/res/id_map.json'
    config.set('General', 'log', 'tsstats/tests/res/test.log')
    assert config.get('General', 'log') == 'tsstats/tests/res/test.log'
    config.set('General', 'output', 'output.html')
    assert config.get('General', 'output') == 'output.html'


def test_read():
    config = load(path='tsstats/tests/res/config.ini')
    # test defaults
    assert not config.getboolean('General', 'debug')
    # test written values
    assert config.get('General', 'log') == 'tsstats/tests/res/test.log'
    assert config.get('General', 'output') == 'tsstats/tests/res/output.html'



# -*- coding: utf-8 -*-
import logging
from collections import namedtuple
from os.path import dirname, join

import pendulum
from jinja2 import ChoiceLoader, Environment, FileSystemLoader, PackageLoader

from tsstats.log import Server
from tsstats.utils import filter_threshold, seconds_to_text, sort_clients

logger = logging.getLogger('tsstats')

SortedClients = namedtuple('SortedClients', [
    'onlinetime', 'kicks', 'pkicks', 'bans', 'pbans'])


def prepare_clients(clients, onlinetime_threshold=-1):
    '''
    Prepare `clients` for rendering

    sort them, clean their nick-history and convert onlinetime to string

    :param clients: List of clients to prepare
    :param onlinetime_threshold: threshold for clients onlinetime

    :type clients: tsstats.client.Clients
    :type onlinetime_treshold: int

    :return: `clients` sorted by onlinetime, kics, pkicks, bans and pbans
    :rtype: tsstats.template.SortedClients
    '''
    # sort by onlinetime
    onlinetime_ = sort_clients(
        clients, lambda c: c.onlinetime.total_seconds()
    )
    # filter clients not matching threshold
    onlinetime_ = filter_threshold(onlinetime_, onlinetime_threshold)
    # convert timespans to text
    onlinetime = [
        (client, seconds_to_text(int(onlinetime)))
        for client, onlinetime in onlinetime_
    ]
    return SortedClients(
        onlinetime=onlinetime,
        kicks=sort_clients(clients, lambda c: c.kicks),
        pkicks=sort_clients(clients, lambda c: c.pkicks),
        bans=sort_clients(clients, lambda c: c.bans),
        pbans=sort_clients(clients, lambda c: c.pbans)
    )


def render_servers(servers, output, title='TeamspeakStats',
                   template='index.jinja2', datetime_fmt='%x %X %Z',
                   onlinetime_threshold=-1, lastseen_relative=True):
    '''
    Render `servers`

    :param servers: list of servers to render
    :param output: path to output-file
    :param template_name: path to template-file
    :param title: title of the resulting html-document
    :param template_path: path to template-file
    :param datetime_fmt: custom datetime-format
    :param onlinetime_threshold: threshold for clients onlinetime
    :param lastseen_relative: render last seen timestamp relative


    :type servers: [tsstats.log.Server]
    :type output: str
    :type template_name: str
    :type title: str
    :type template_path: str
    :type datetime_fmt: str
    :type onlinetime_threshold: int
    :type lastseen_relative: bool
    '''
    # preparse servers
    prepared_servers = [
        Server(sid, prepare_clients(clients, onlinetime_threshold))
        for sid, clients in servers
    ]
    # render
    template_loader = ChoiceLoader([
        PackageLoader(__package__, 'templates'),
        FileSystemLoader(join(dirname(__file__), 'templates'))
    ])
    template_env = Environment(loader=template_loader)

    def frmttime(timestamp):
        if not timestamp:
            return ''
        formatted = timestamp.strftime(datetime_fmt)
        logger.debug('Formatting timestamp %s -> %s', timestamp, formatted)
        return formatted

    def lastseen(timestamp):
        if lastseen_relative:
            return timestamp.diff_for_humans()
        else:
            return frmttime(timestamp)
    template_env.filters['frmttime'] = frmttime
    template_env.filters['lastseen'] = lastseen
    template = template_env.get_template(template)
    logger.debug('Rendering template %s', template)
    template.stream(title=title, servers=prepared_servers,
                    debug=logger.level <= logging.DEBUG,
                    creation_time=pendulum.now('UTC'))\
        .dump(output, encoding='utf-8')
    logger.debug('Wrote rendered template to %s', output)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['tsstats', 'tsstats.tests']

package_data = \
{'': ['*'], 'tsstats': ['templates/*'], 'tsstats.tests': ['res/*']}

install_requires = \
['jinja2>=3.1.2,<4.0.0', 'pendulum>=2.1.2,<3.0.0']

entry_points = \
{'console_scripts': ['tsstats = tsstats.__main__:cli']}

setup_kwargs = {
    'name': 'tsstats',
    'version': '2.1.0',
    'description': 'A simple Teamspeak stats generator',
    'long_description': "TeamspeakStats |Build Status| |Build status| |Coverage Status| |PyPI| |Documentation Status|\n==========================================================================================================\n\nA simple Teamspeak stat-generator - based solely on server-logs\n\n|screenshot|\n\nInstallation\n============\n\n-  Install the package via PyPi ``pip install tsstats``\n-  Clone this repo\n   ``git clone https://github.com/Thor77/TeamspeakStats`` and install\n   with ``python setup.py install``\n-  Just use the package as is via ``python -m tsstats [-h]``\n\nUsage\n=====\n\n-  Run the script ``tsstats [-h]``\n-  Optionally create a config-file (see\n   `Configuration <https://teamspeakstats.readthedocs.io/en/latest/config.html>`__)\n-  The package works entirely off your Teamspeak server's logs, so that\n   no ServerQuery account is necessary\n\nExample\n=======\n\n::\n\n    tsstats -l /var/log/teamspeak3-server/ -o /var/www/tsstats.html\n\nParse logs in ``/var/log/teamspeak3-server`` and write output to ``/var/www/tsstats.html``.\n\nFor more details checkout the `documentation <http://teamspeakstats.readthedocs.io/en/latest/>`__!\n\n.. |screenshot| image:: https://raw.githubusercontent.com/Thor77/TeamspeakStats/master/screenshot.png\n.. |Build Status| image:: https://travis-ci.org/Thor77/TeamspeakStats.svg?branch=master\n   :target: https://travis-ci.org/Thor77/TeamspeakStats\n.. |Build status| image:: https://ci.appveyor.com/api/projects/status/u9cx7krwmmevbvl2/branch/master?svg=true\n   :target: https://ci.appveyor.com/project/Thor77/teamspeakstats\n.. |Coverage Status| image:: https://coveralls.io/repos/Thor77/TeamspeakStats/badge.svg?branch=master&service=github\n   :target: https://coveralls.io/github/Thor77/TeamspeakStats?branch=master\n.. |PyPI| image:: https://img.shields.io/pypi/v/tsstats.svg\n   :target: https://pypi.python.org/pypi/tsstats\n.. |Documentation Status| image:: https://readthedocs.org/projects/teamspeakstats/badge/?version=latest\n   :target: http://teamspeakstats.readthedocs.io/en/latest/?badge=latest\n",
    'author': 'Thor77',
    'author_email': 'thor77@thor77.org',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



# -*- coding: utf-8 -*-
import pendulum
import pytest

from tsstats import events
from tsstats.log import TimedLog, _bundle_logs, _parse_line, parse_logs
from tsstats.template import render_servers

testlog_path = 'tsstats/tests/res/test.log'

static_timestamp = pendulum.datetime(2015, 5, 18, 15, 52, 52, 685612)


@pytest.fixture
def clients():
    return list(parse_logs(testlog_path, online_dc=False))[0].clients


def test_log_client_count(clients):
    assert len(clients) == 3


def test_log_onlinetime(clients):
    assert clients['1'].onlinetime == pendulum.duration(
        seconds=402, microseconds=149208)
    assert clients['2'].onlinetime == pendulum.duration(
        seconds=19, microseconds=759644)


def test_log_kicks(clients):
    assert clients['UIDClient1'].kicks == 1


def test_log_pkicks(clients):
    assert clients['2'].pkicks == 1


def test_log_bans(clients):
    assert clients['UIDClient1'].bans == 1


def test_log_pbans(clients):
    assert clients['2'].pbans == 1


@pytest.mark.parametrize("logs,bundled", [
    (
        ['l1.log', 'l2.log'],
        {'': [TimedLog('l1.log', None), TimedLog('l2.log', None)]}
    ),
    (
        [
            'ts3server_2016-06-06__14_22_09.527229_1.log',
            'ts3server_2017-07-07__15_23_10.638340_1.log'
        ],
        {
            '1': [
                TimedLog(
                    'ts3server_2016-06-06__14_22_09.527229_1.log',
                    pendulum.datetime(
                        year=2016, month=6, day=6, hour=14, minute=22,
                        second=9, microsecond=527229
                    )
                ),
                TimedLog(
                    'ts3server_2017-07-07__15_23_10.638340_1.log',
                    pendulum.datetime(
                        year=2017, month=7, day=7, hour=15, minute=23,
                        second=10, microsecond=638340
                    )
                )
            ]
        }
    )
])
def test_log_bundle(logs, bundled):
    assert _bundle_logs(logs) == bundled


def test_log_client_online():
    current_time = pendulum.now()

    pendulum.set_test_now(current_time)
    clients = list(parse_logs(testlog_path))[0].clients
    old_onlinetime = int(clients['1'].onlinetime.total_seconds())

    pendulum.set_test_now(current_time.add(seconds=2))  # add 2s to .now()
    clients = list(parse_logs(testlog_path))[0].clients
    assert int(clients['1'].onlinetime.total_seconds()) == old_onlinetime + 2


def test_parse_groups():
    server = list(parse_logs('tsstats/tests/res/test.log.groups'))
    assert len(server) == 0


def test_parse_utf8(output):
    servers = parse_logs(testlog_path + '.utf8')
    render_servers(servers, output)


def test_parse_invalid_line():
    assert _parse_line('INVALID') == []


@pytest.mark.parametrize('line,expected_events', [
    (
        "client connected 'Client1'(id:1) from 1.2.3.4:1234",
        [
            events.connect(static_timestamp, '1')
        ]
    ),
    (
        "client disconnected 'Client1'(id:1) reason 'reasonmsg=ByeBye!'",
        [
            events.disconnect(static_timestamp, '1')
        ]
    ),
    (
        "client disconnected 'Client1'(id:1) reason 'invokerid=1"
        " invokername=Client2 invokeruid=UIDClient2 reasonmsg'",
        [
            events.disconnect(static_timestamp, '1'),
            events.nick(None, 'UIDClient2', 'Client2'),
            events.kick(None, 'UIDClient2', '1')
        ]
    ),
    (
        "client disconnected 'Client1'(id:1) reason 'invokerid=2 "
        "invokername=Client2 invokeruid=UIDClient2 reasonmsg bantime=0'",
        [
            events.disconnect(static_timestamp, '1'),
            events.nick(None, 'UIDClient2', 'Client2'),
            events.ban(None, 'UIDClient2', '1')
        ]
    )
])
def test_parse_line(line, expected_events):
    line = '2015-05-18 15:52:52.685612|INFO    |VirtualServerBase|  3| ' + line
    expected_events.insert(0, events.nick(None, '1', 'Client1'))
    expected_events = [
        event._replace(timestamp=static_timestamp) for event in expected_events
    ]
    assert _parse_line(line) == expected_events



# -*- coding: utf-8 -*-
import pytest

from tsstats.client import Client, Clients


@pytest.fixture(scope='module')
def clients():
    clients = Clients()
    cl1 = Client('1')
    cl2 = Client('2')
    clients += cl1
    clients += cl2
    uidcl1 = Client('UID1')
    uidcl2 = Client('UID2')
    clients += uidcl1
    clients += uidcl2
    return (clients, cl1, cl2, uidcl1, uidcl2)


def test_client_get(clients):
    clients, cl1, cl2, uidcl1, uidcl2 = clients
    assert clients['1'] == cl1
    assert clients['2'] == cl2
    assert clients['UID1'] == uidcl1
    assert clients['UID2'] == uidcl2
    with pytest.raises(KeyError):
        clients['3']
        clients['UID3']


def test_client_repr(clients):
    clients, _, _, _, _ = clients
    assert str(clients['1']) == '<1, None>'
    assert str(clients['2']) == '<2, None>'
    assert str(clients['UID1']) == '<UID1, None>'
    assert str(clients['UID2']) == '<UID2, None>'
    assert repr(clients['1']) == str(clients['1'])


def test_client_nick(clients):
    _, cl1, _, _, _ = clients
    assert cl1.nick is None
    assert not cl1.nick_history
    cl1.nick = 'Client1'
    assert cl1.nick == 'Client1'
    assert None not in cl1.nick_history
    cl1.nick = 'NewClient1'
    assert cl1.nick == 'NewClient1'
    assert 'Client1' in cl1.nick_history


def test_clients_iter(clients):
    clients, cl1, cl2, uidcl1, uidcl2 = clients
    client_list = list(clients.values())
    assert cl1 in client_list
    assert cl2 in client_list
    assert uidcl1 in client_list
    assert uidcl2 in client_list


def test_clients_delete(clients):
    clients, cl1, _, _, _ = clients
    del clients['1']
    assert '1' not in clients
